""" Test pythia.basis module. """
from unittest.mock import MagicMock
import numpy as np
from scipy.special import gamma
from scipy.integrate import quad

import pythia.sampler as sampler


def test_UniformSampler() -> None:
    """Test uniform sampler."""

    domain = [-1, 2]
    diam = domain[1] - domain[0]
    s = sampler.UniformSampler(domain)

    # test general properties
    assert isinstance(s, sampler.Sampler)
    assert isinstance(s.domain, np.ndarray)
    assert s.dimension == 1
    assert s.mass == 1
    assert np.abs(s.maximum - 1 / diam) < 1e-12
    assert np.abs(s.mean - 0.5) < 1e-12
    assert np.abs(s.var - 3/4) < 1e-12
    assert np.abs(s.std - np.sqrt(3)/2) < 1e-12

    # test pdf
    vals = s.pdf(np.random.uniform(*domain, 100))
    assert vals.shape == (100,)
    assert np.all(vals - 1/diam) < 1e-12

    # test log-pdf
    vals = s.log_pdf(np.random.uniform(*domain, 100))
    assert vals.shape == (100,)
    assert np.all(vals - np.log(1/diam)) < 1e-12

    # test grad log-pdf
    vals = s.grad_x_log_pdf(np.random.uniform(*domain, 100))
    assert vals.shape == (100,)
    assert np.all(vals) < 1e-12

    # test hessian log-pdf
    vals = s.hess_x_log_pdf(np.random.uniform(*domain, 100))
    assert vals.shape == (100,)
    assert np.all(vals) < 1e-12

    # test sampling
    # Note: There is no reasonable way to reliably test if moments of
    #       distribution are correct (without drawing a lot of samples)
    samples = s.sample((100, 1))
    assert samples.shape == (100, 1)
    assert np.all(samples > domain[0]) and np.all(samples < domain[1])


def test_NormalSampler() -> None:
    """Test normal sampler."""

    mean = 1
    var = 4
    s = sampler.NormalSampler(mean, var)

    # test general properties
    assert isinstance(s, sampler.Sampler)
    assert isinstance(s.domain, np.ndarray)
    assert np.abs(s.mean - 1) < 1e-12
    assert np.abs(s.var - 4) < 1e-12
    assert np.abs(s.std - 2) < 1e-12
    assert s.dimension == 1
    assert s.mass == 1
    assert np.abs(s.maximum - 1 / np.sqrt(2*np.pi*var)) < 1e-12

    # test pdf
    xs = np.random.normal(mean, np.sqrt(var), 100)
    vals = s.pdf(xs)
    expected = np.exp(-0.5 * (xs-mean)**2 / var) / np.sqrt(2*np.pi*var)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test log-pdf
    xs = np.random.normal(mean, np.sqrt(var), 100)
    vals = s.log_pdf(xs)
    expected = -0.5 * (xs-mean)**2 / var - np.log(np.sqrt(2*np.pi*var))
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test grad log-pdf
    xs = np.random.normal(mean, np.sqrt(var), 100)
    vals = s.grad_x_log_pdf(xs)
    expected = - (xs-mean) / var
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test hessian log-pdf
    xs = np.random.normal(mean, np.sqrt(var), 100)
    vals = s.hess_x_log_pdf(xs)
    expected = - 1 / var
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test sampling
    # Note: There is no reasonable way to reliably test if moments of
    #       distribution are correct (without drawing a lot of samples)
    assert s.sample((100, 1)).shape == (100, 1)


def test_GammaSampler() -> None:
    """Test gamma sampler."""

    domain = [1, np.inf]
    alpha = 1
    beta = 3
    s = sampler.GammaSampler(domain, alpha, beta)

    # test general properties
    assert isinstance(s, sampler.Sampler)
    assert isinstance(s.domain, np.ndarray)
    assert s.dimension == 1
    assert s.mass == 1
    assert s.alpha == 1
    assert s.beta == 3
    assert np.abs(s.mean - 4/3) < 1e-12
    assert np.abs(s.var - 1/9) < 1e-12
    assert np.abs(s.std - 1/3) < 1e-12
    assert np.abs(s.maximum - 3) < 1e-12

    # test pdf
    xs = np.random.gamma(alpha, 1/beta, 100) + domain[0]
    vals = s.pdf(xs)
    expected = 3 * np.exp(-3*(xs-domain[0]))  # pdf for alpha=1 and beta=3
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test log-pdf
    xs = np.random.gamma(alpha, 1/beta, 100) + domain[0]
    vals = s.log_pdf(xs)
    expected = np.log(3) - 3*(xs-domain[0])  # log-pdf for alpha=1 and beta=3
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test grad log-pdf
    # Note: not yet implemented.

    # test hessian log-pdf
    # Note: not yet implemented.

    # test sampling
    # Note: There is no reasonable way to reliably test if moments of
    #       distribution are correct (without drawing a lot of samples)
    assert s.sample((100, 1)).shape == (100, 1)


def test_BetaSampler() -> None:
    """Test beta sampler."""

    domain = [1, 5]
    length = domain[1]-domain[0]
    alpha = 2
    beta = 2
    s = sampler.BetaSampler(domain, alpha, beta)

    # test general properties
    assert isinstance(s, sampler.Sampler)
    assert isinstance(s.domain, np.ndarray)
    assert s.dimension == 1
    assert s.mass == 1
    assert s.alpha == 2
    assert s.beta == 2
    assert np.abs(s.mean - 3) < 1e-12
    assert np.abs(s.var - 1/5) < 1e-12
    assert np.abs(s.std - 1/np.sqrt(5)) < 1e-12
    assert np.abs(s.maximum - 6/16) < 1e-12

    # test pdf
    Beta = gamma(4) / (gamma(2)*gamma(2))
    xs = np.random.beta(alpha, beta, 100)*length + domain[0]
    vals = s.pdf(xs)
    # pdf for alpha=2 and beta=2
    expected = length**(-3) * Beta * (xs-domain[0]) * (domain[1] - xs)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test log-pdf
    xs = np.random.beta(alpha, beta, 100)*length + domain[0]
    vals = s.log_pdf(xs)
    expected = (-3*np.log(length) + np.log(Beta)
                + np.log(xs-domain[0]) + np.log(domain[1]-xs))
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test grad log-pdf
    # Note: not yet implemented.

    # test hessian log-pdf
    # Note: not yet implemented.

    # test sampling
    # Note: There is no reasonable way to reliably test if moments of
    #       distribution are correct (without drawing a lot of samples)
    assert s.sample((100, 1)).shape == (100, 1)


def test_WLSUnivariateSampler() -> None:
    """Test univariate wls sampler.

    .. note::
        This currently only tests uniform WLS sampling.
    """

    param = MagicMock()
    param.distribution = "uniform"
    param.domain = [-1, 1]
    s = sampler.WLSUnivariateSampler(param, deg=5, tsa=True)
    # define normalized Legendre basis polynomials
    basis = [np.polynomial.legendre.Legendre([0]*j+[1], param.domain)
             for j in range(6)]
    scales = [quad(lambda x: 0.5*p(x)**2, param.domain[0], param.domain[1])[0]
              for p in basis]
    basis = [p / np.sqrt(scale) for (p, scale) in zip(basis, scales)]

    # test general properties
    assert isinstance(s, sampler.Sampler)
    assert isinstance(s.domain, np.ndarray)
    assert s.dimension == 1
    assert s.mass == 1
    assert np.abs(s.mean - 0.0) < 1e-10
    assert np.abs(s.var - 0.4965034965034964) < 1e-10
    assert np.abs(s.std - np.sqrt(0.4965034965034964)) < 1e-10
    assert np.abs(s.maximum - 3) < 1e-10

    # test weight
    xs = np.random.uniform(param.domain[0], param.domain[1], 100)
    expected = 6 / np.sum([np.abs(p(xs))**2 for p in basis], axis=0)
    vals = s.weight(xs)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test pdf
    xs = np.random.uniform(param.domain[0], param.domain[1], 100)
    vals = s.pdf(xs)
    expected = 1/12 * np.sum([np.abs(p(xs))**2 for p in basis], axis=0)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test log-pdf
    xs = np.random.uniform(param.domain[0], param.domain[1], 100)
    vals = s.log_pdf(xs)
    expected = np.log(np.sum([np.abs(p(xs))**2 for p in basis], axis=0)/12)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test grad log-pdf
    # Note: not yet implemented.

    # test hessian log-pdf
    # Note: not yet implemented.

    # test _compute_trial_sampler
    xs = np.random.uniform(param.domain[0], param.domain[1], 100)
    trial_sampler, bulk = s._compute_trial_sampler()
    assert np.all(bulk*trial_sampler.pdf(xs) > s.pdf(xs))

    # test sampling
    # Note: There is no reasonable way to reliably test if moments of
    #       distribution are correct (without drawing a lot of samples)
    assert s.sample((100, 1)).shape == (100, 1)


def test_ProductSampler() -> None:
    """Test product sampler."""

    # NOTE: Mock samplers do not resemble real univariate samplers.
    sampler_1 = MagicMock()
    sampler_1.domain = np.array([[-1, 3]])
    sampler_1.maximum = 1/4
    sampler_1.mean = 1.0
    sampler_1.cov = 4/3
    sampler_1.pdf = lambda x: 1/4*np.ones(x.shape[0])
    sampler_1.log_pdf = lambda x: np.log(1/4)*np.ones(x.size)
    sampler_1.grad_x_log_pdf = lambda x: 1/4*np.ones(x.size)
    sampler_1.hess_x_log_pdf = lambda x: 1/4*np.ones(x.size)
    sampler_1.sample = lambda shape: np.ones(shape)

    sampler_2 = MagicMock()
    sampler_2.domain = np.array([[-np.inf, np.inf]])
    sampler_2.maximum = 1 / np.sqrt(4*np.pi)
    sampler_2.mean = 2.0
    sampler_2.cov = 2.0
    sampler_2.pdf = lambda x: np.exp(-x**2/2)
    sampler_2.log_pdf = lambda x: -x**2/2
    sampler_2.grad_x_log_pdf = lambda x: -x
    sampler_2.hess_x_log_pdf = lambda x: -x
    sampler_2.sample = lambda shape: np.ones(shape)

    s = sampler.ProductSampler([sampler_1, sampler_2])

    # test general properties
    assert isinstance(s, sampler.Sampler)
    assert isinstance(s.domain, np.ndarray)
    assert np.linalg.norm(s.domain[0] - np.array([-1, 3])) < 1e-12
    assert np.all(np.isinf(s.domain[1]))
    assert s.dimension == 2
    assert s.mass == 1
    assert np.abs(s.maximum - 1 / (8*np.sqrt(np.pi))) < 1e-12
    assert np.linalg.norm(s.mean - [1, 2]) < 1e-12
    assert np.linalg.norm(s.cov - np.diag([4/3, 2])) < 1e-12

    # test pdf
    xs = np.random.uniform(-1, 3, (100, 2))
    expected = 1/4 * np.exp(-xs[:, 1]**2/2)
    assert s.pdf(xs).shape == (100,)
    assert np.max(np.abs(s.pdf(xs) - expected)) < 1e-12

    # test log-pdf
    xs = np.random.uniform(-1, 3, (100, 2))
    expected = np.log(1/4) - xs[:, 1]**2/2
    assert s.log_pdf(xs).shape == (100,)
    assert np.max(np.abs(s.log_pdf(xs) - expected)) < 1e-12

    # test grad log-pdf
    xs = np.random.uniform(-1, 3, (100, 2))
    expected = np.array([1/4*np.ones(xs.shape[0]), - xs[:, 1]]).T
    assert s.grad_x_log_pdf(xs).shape == (100, 2)
    assert np.max(np.abs(s.grad_x_log_pdf(xs) - expected)) < 1e-12

    # test hessian log-pdf
    xs = np.random.uniform(-1, 3, (100, 2))
    expected = np.zeros((100, 2, 2))
    expected[:, 0, 0] = 1/4*np.ones(xs.shape[0])
    expected[:, 1, 1] = - xs[:, 1]
    assert s.hess_x_log_pdf(xs).shape == (100, 2, 2)
    assert np.max(np.abs(s.hess_x_log_pdf(xs) - expected)) < 1e-12

    # test sampling
    # Note: There is no reasonable way to reliably test if moments of
    #       distribution are correct (without drawing a lot of samples)
    assert np.linalg.norm(s.sample((100, 1)) - np.ones((100, 1, 2))) < 1e-12


def test_ParameterSampler() -> None:
    """Test parameter sampler."""
    param1 = MagicMock()
    param1.distribution = "normal"
    param1.domain = [-np.inf, np.inf]
    param1.mean = 0.0
    param1.var = 1.0

    param2 = MagicMock()
    param2.distribution = "normal"
    param2.domain = [-np.inf, np.inf]
    param2.mean = 1.0
    param2.var = 2.0

    s = sampler.ParameterSampler([param1, param2])

    # test general properties
    assert isinstance(s, sampler.Sampler)
    assert isinstance(s.domain, np.ndarray)
    assert np.all(s.domain.shape == (2, 2))
    assert np.all(np.isinf(s.domain))
    assert s.dimension == 2
    assert s.mass == 1
    assert np.abs(s.maximum - 1 / (np.sqrt(2)*2*np.pi)) < 1e-12
    assert np.linalg.norm(s.mean - [0, 1]) < 1e-12
    assert np.linalg.norm(s.cov - np.diag([1, 2])) < 1e-12

    # test pdf
    xs = np.random.multivariate_normal(s.mean, s.cov, 100)
    expected = (1/(np.sqrt(2)*2*np.pi)
                * np.exp(-xs[:, 0]**2/2 - (xs[:, 1]-1)**2/4))
    assert s.pdf(xs).shape == (100,)
    assert np.max(np.abs(s.pdf(xs) - expected)) < 1e-12

    # test log-pdf
    xs = np.random.multivariate_normal(s.mean, s.cov, 100)
    expected = (-np.log((np.sqrt(2)*2*np.pi))
                - xs[:, 0]**2/2 - (xs[:, 1]-1)**2/4)
    assert s.log_pdf(xs).shape == (100,)
    assert np.max(np.abs(s.log_pdf(xs) - expected)) < 1e-12

    # test grad log-pdf
    xs = np.random.multivariate_normal(s.mean, s.cov, 100)
    expected = np.array([-xs[:, 0], -(xs[:, 1]-1)/2]).T
    assert s.grad_x_log_pdf(xs).shape == (100, 2)
    assert np.max(np.abs(s.grad_x_log_pdf(xs) - expected)) < 1e-12

    # test hessian log-pdf
    xs = np.random.multivariate_normal(s.mean, s.cov, 100)
    expected = np.zeros((100, 2, 2))
    expected[:, 0, 0] = -np.ones(xs.shape[0])
    expected[:, 1, 1] = -np.ones(xs.shape[0])/2.0
    assert s.hess_x_log_pdf(xs).shape == (100, 2, 2)
    assert np.max(np.abs(s.hess_x_log_pdf(xs) - expected)) < 1e-12

    # test sampling
    # Note: There is no reasonable way to reliably test if moments of
    #       distribution are correct (without drawing a lot of samples)
    assert s.sample((100, 1)).shape == (100, 1, 2)


def test_WLSSampler() -> None:
    """Test weighted least-squares sampler."""
    param = MagicMock()
    param.distribution = "uniform"
    param.domain = [-1, 1]

    # define normalized Legendre basis polynomials
    basis = [np.polynomial.legendre.Legendre([0]*j+[1], param.domain)
             for j in range(6)]
    scales = [quad(lambda x: 0.5*p(x)**2, param.domain[0], param.domain[1])[0]
              for p in basis]
    basis = [p / np.sqrt(scale) for (p, scale) in zip(basis, scales)]
    m_basis = []
    for p in basis:
        for q in basis:
            def func(x, p=p, q=q):
                return p(x[:, 0])*q(x[:, 1])
            m_basis.append(func)
    s = sampler.WLSSampler([param, param], m_basis, tsa=True)

    # test general properties
    assert isinstance(s, sampler.Sampler)
    assert isinstance(s.domain, np.ndarray)
    assert s.dimension == 2
    assert s.mass == 1
    assert np.abs(s.maximum - 9) < 1e-10

    # Note: mean and cov not implemented

    # test weight
    xs = np.random.uniform(param.domain[0], param.domain[1], (100, 2))
    expected = (6**2 / np.prod(np.sum(
        [np.abs(p(xs))**2 for p in basis], axis=0), axis=1))
    vals = s.weight(xs)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test pdf
    xs = np.random.uniform(param.domain[0], param.domain[1], (100, 2))
    vals = s.pdf(xs)
    expected = 1/12**2 * np.prod(np.sum(
        [np.abs(p(xs))**2 for p in basis], axis=0), axis=1)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test log-pdf
    xs = np.random.uniform(param.domain[0], param.domain[1], (100, 2))
    vals = s.log_pdf(xs)
    expected = np.sum(np.log(np.sum(
        [np.abs(p(xs))**2 for p in basis], axis=0)), axis=1) - np.log(12**2)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test grad log-pdf
    # Note: not yet implemented.

    # test hessian log-pdf
    # Note: not yet implemented.

    # test sampling
    # Note: There is no reasonable way to reliably test if moments of
    #       distribution are correct (without drawing a lot of samples)
    assert s.sample((100, 1)).shape == (100, 1, 2)


def test_WLSTensorSampler() -> None:
    """Test weighted least-squares tensor product sampler."""
    param = MagicMock()
    param.distribution = "uniform"
    param.domain = [-1, 1]
    s = sampler.WLSTensorSampler([param, param], [5, 5], tsa=True)

    # define normalized Legendre basis polynomials
    basis = [np.polynomial.legendre.Legendre([0]*j+[1], param.domain)
             for j in range(6)]
    scales = [quad(lambda x: 0.5*p(x)**2, param.domain[0], param.domain[1])[0]
              for p in basis]
    basis = [p / np.sqrt(scale) for (p, scale) in zip(basis, scales)]

    # test general properties
    assert isinstance(s, sampler.Sampler)
    assert isinstance(s.domain, np.ndarray)
    assert s.dimension == 2
    assert s.mass == 1
    assert np.abs(s.maximum - 9) < 1e-10
    assert np.linalg.norm(s.mean - [0, 0]) < 1e-12
    var = 0.4965034965034964
    assert np.linalg.norm(s.cov - np.diag([var, var])) < 1e-12

    # test weight
    xs = np.random.uniform(param.domain[0], param.domain[1], (100, 2))
    expected = (6**2 / np.prod(np.sum(
        [np.abs(p(xs))**2 for p in basis], axis=0), axis=1))
    vals = s.weight(xs)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test pdf
    xs = np.random.uniform(param.domain[0], param.domain[1], (100, 2))
    vals = s.pdf(xs)
    expected = 1/12**2 * np.prod(np.sum(
        [np.abs(p(xs))**2 for p in basis], axis=0), axis=1)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test log-pdf
    xs = np.random.uniform(param.domain[0], param.domain[1], (100, 2))
    vals = s.log_pdf(xs)
    expected = np.sum(np.log(np.sum(
        [np.abs(p(xs))**2 for p in basis], axis=0)), axis=1) - np.log(12**2)
    assert vals.shape == (100,)
    assert np.max(np.abs(vals - expected)) < 1e-12

    # test grad log-pdf
    # Note: not yet implemented.

    # test hessian log-pdf
    # Note: not yet implemented.

    # test sampling
    # Note: There is no reasonable way to reliably test if moments of
    #       distribution are correct (without drawing a lot of samples)
    assert s.sample((100, 1)).shape == (100, 1, 2)


def test_rejection_sampling() -> None:
    """Test rejection sampling."""
    # TODO(Nando): Is there a reasonable way to test rejection sampling?
    ts = MagicMock()
    ts.pdf = lambda x: np.ones(x.shape[0])
    ts.sample = lambda n: np.random.uniform(0, 1, (n, 2))
    samples = sampler.rejection_sampling(
        ts.pdf, ts, scale=1.1, dimension=2, shape=(1000, 3))
    assert samples.shape == (1000, 3, 2)


def test_constraint_sampling() -> None:
    """Test constraint sampling."""
    s = MagicMock()
    s.sample = lambda n: np.random.uniform(0, 1, (np.prod(n), 2))
    s.dimension = 2

    def constraint(x):
        if x.ndim != 1:
            x = x.reshape(-1)
        if x[0] + x[1] <= 1:
            return True
        return False

    # test sampling shape and constraint
    samples = sampler.constraint_sampling(s, [constraint], 1000)
    assert samples.shape == (1000, 2)
    assert np.all(np.sum(samples, axis=1) <= 1)


def test_assign_sampler() -> None:
    """Test assigning of correct sampler function."""
    param = MagicMock()
    param.distribution = "uniform"
    param.domain = [0, 1]
    assert isinstance(sampler.assign_sampler(param), sampler.UniformSampler)

    param = MagicMock()
    param.distribution = "normal"
    param.mean = 1
    param.var = 1
    assert isinstance(sampler.assign_sampler(param), sampler.NormalSampler)

    param = MagicMock()
    param.distribution = "gamma"
    param.domain = [1, np.inf]
    param.alpha = 2
    param.beta = 0.5
    assert isinstance(sampler.assign_sampler(param), sampler.GammaSampler)

    param = MagicMock()
    param.distribution = "beta"
    param.domain = [0, 1]
    param.alpha = 2
    param.beta = 0.5
    assert isinstance(sampler.assign_sampler(param), sampler.BetaSampler)


def test_get_maximum() -> None:
    """Test estimate of function maximum."""
    # maximum at border of domain
    val = sampler.get_maximum(
        lambda x: x[:, 0]**2 * x[:, 1]**2, [[-1, 1], [-1, 1]])
    assert np.linalg.norm(val - 1) < 1e-2

    # maximum in domain
    mean = np.random.uniform(-0.5, 0.5, (1, 2))
    val = sampler.get_maximum(
        lambda x: np.exp(- np.linalg.norm(x-mean, axis=1)**2/2),
        [[-1, 1], [-1, 1]])
    assert np.linalg.norm(val-1) < 1e-2


if __name__ == "__main__":
    pass



""" Test pythia.parameter module. """
import pytest
import numpy as np
import pythia.parameter as parameter


def test_parameter() -> None:
    """Test Parameter class."""
    param1 = parameter.Parameter(0, "uni", [0, 1], "uniform")
    assert isinstance(param1, parameter.Parameter)
    assert param1.name == "uni"
    assert param1.index == 0
    assert np.all(param1.domain == np.array([0, 1]))
    assert param1.distribution == "uniform"
    assert param1.mean is None
    assert param1.var is None
    assert param1.alpha is None
    assert param1.beta is None

    param2 = parameter.Parameter(1, "nor", [-np.inf, np.inf], "normal",
                                 mean=0, var=1)
    assert param2.mean == 0.0
    assert param2.var == 1.0
    assert param2.alpha is None
    assert param2.beta is None

    param3 = parameter.Parameter(2, "beta", [-1, 1], "beta",
                                 alpha=1, beta=1)
    assert param3.mean is None
    assert param3.var is None
    assert param3.alpha == 1
    assert param3.beta == 1

    with pytest.raises(ValueError):  # bernoulli is unsupported distribution
        _ = parameter.Parameter(-1, "_", [-1, 1], "bernoulli", mean=0, var=-1)

    with pytest.raises(AssertionError):  # mean is not None
        _ = parameter.Parameter(-1, "_", [-1, 1], "uniform", mean=0)

    with pytest.raises(AssertionError):  # alpha is not None
        _ = parameter.Parameter(-1, "_", [-np.inf, np.inf], "normal",
                                mean=0, var=1, alpha=1)

    with pytest.raises(AssertionError):  # var is missing
        _ = parameter.Parameter(-1, "_", [-np.inf, np.inf], "normal", mean=0)

    with pytest.raises(AssertionError):  # var is smaller then zero
        _ = parameter.Parameter(-1, "_", [-np.inf, np.inf], "normal",
                                mean=0, var=-1)

    with pytest.raises(AssertionError):  # mean is not None
        _ = parameter.Parameter(-1, "_", [-1, 1], "beta",
                                mean=0, alpha=1, beta=1)

    with pytest.raises(AssertionError):  # alpha is missing
        _ = parameter.Parameter(-1, "_", [-1, 1], "beta", beta=1)



""" Test pythia.misc module. """
import pytest
import numpy as np
import pythia.misc as misc


def test_distributionDict() -> None:
    """Test distribution dictionary mappings.
    """
    dist_dict = misc.distributionDict()
    assert all(dist in ['uniform', 'Uniform', 'U']
               for dist in dist_dict["uniform"])
    assert all(dist in ['normal', 'gauss', 'Gauss', 'gaussian', 'N']
               for dist in dist_dict["normal"])
    assert all(dist in ['gamma', 'Gamma', 'G']
               for dist in dist_dict["gamma"])
    assert all(dist in ['beta', 'Beta', 'B']
               for dist in dist_dict["beta"])


def test_shiftCoord() -> None:
    """Test affine transformation from one interval to another."""

    # normal usecase float input
    assert misc.shiftCoord(0, [-1, 1], [0, 1]) == 0.5

    # normal usecase array input
    points = np.random.uniform(0, 1, (100))
    assert np.all(misc.shiftCoord(points, [0, 1], [1, 3]) == 2*points+1)

    # nothing happens if both intervals are the same
    assert misc.shiftCoord(0, [-1, 1], [-1, 1]) == 0

    # shifting values outside of original interval
    points = np.array([-2, 2])
    assert np.all(misc.shiftCoord(points, [0, 1], [1, 3]) == [-3, 5])


def test_cardProd() -> None:
    """Test build of cartesian product."""
    # empty list raises error
    with pytest.raises(ValueError):
        _ = misc.cartProd([]) == []

    # one array in list returns reshaped array
    array = np.arange(10)
    assert np.all(misc.cartProd([array]) == array.reshape(-1, 1))

    # only one point in each dimension
    array = [np.array([j]) for j in range(10)]
    assert np.all(misc.cartProd(array) == np.arange(10).reshape(1, -1))

    # single array is treated as n-D list with one point each
    # (same result as previous test)
    array = np.arange(10)
    assert np.all(misc.cartProd(array) == array.reshape(1, -1))

    # multiple arrays of different length (standard usecase)
    result = np.array([[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2]])
    assert np.all(misc.cartProd([np.arange(2), np.arange(3)]) == result)


def test_is_containted() -> None:
    """Test containment of points in domain."""

    # normal usecase 1-dim input (float or int)
    assert misc.is_contained(0.5, [0, 1]) is True  # point in domain
    assert misc.is_contained(0, [0, 1]) is True  # left boundary of domain
    assert misc.is_contained(1, [0, 1]) is True  # right boundary of domain
    assert misc.is_contained(-1, [0, 1]) is False  # point outside domain

    # normal usecase n-dim input (array, list or tuple)
    domain = [[0, 1], [0, 1], [0, 1]]
    assert misc.is_contained(0.5*np.ones(3), domain) is True  # in domain
    assert misc.is_contained([0, 0, 0], domain) is True  # boundary of domain
    assert misc.is_contained((2, 0, 0), domain) is False  # outside of domain


def test_formatTime() -> None:
    """Test correct formatting of time for printing."""

    # negative times cause assertion error
    with pytest.raises(AssertionError):
        _ = misc.formatTime(-1)

    # zero returns zero
    assert misc.formatTime(0) == "0s"

    # result is rounded to next full seconds
    assert misc.formatTime(1.2) == "2s"

    # normal usecase
    assert misc.formatTime(1234567) == "14 days 6h 56min 8s"


def test_now() -> None:
    """Test now formatting string.

    .. note::
        Testing is not necessary for a simple time formatting string.
    """
    pass


def test_line() -> None:
    """Test printing of separator line."""

    # empty indicator raises error
    with pytest.raises(AssertionError):
        _ = misc.line("")

    # normal line is 80 characters wide
    assert misc.line("-") == "-"*80

    # repeat indicator sequence until 80 characters reached
    assert misc.line("-+#") == "-+#"*26 + "-+"

    # added label doesn't change line length of 80 characters
    assert misc.line("-", "PyThia") == "-- PyThia " + "-"*70


def test_save() -> None:
    """Test save wrapper.

    .. note::
        Testing is not necessary as this is a wrapper for ``numpy.save()``.
    """
    pass


def test_load() -> None:
    """Test load wrapper.

    .. note::
        Testing is not necessary as this is a wrapper for ``numpy.load()``.
    """
    pass


def test_str2iter() -> None:
    """Test casting of str(list) to list."""

    # empty string raises error
    assert misc.str2iter("", list, int) == []
    assert misc.str2iter("", tuple, int) == ()

    # cast string representation of list to iterable
    assert misc.str2iter("[1,2,3]", list, int) == [1, 2, 3]

    # spaces are irrelevant
    assert misc.str2iter("[1, 2, 3]", list, int) == [1, 2, 3]

    # one element is cast correctly
    assert misc.str2iter("[1]", tuple, float) == (1.0,)

    # casting from tuple is ok
    assert misc.str2iter("(1, 2, 3)", list, int) == [1, 2, 3]

    # using trailing ',' is ignored
    assert misc.str2iter("(1,)", list, int) == [1]
    assert misc.str2iter("(1, )", list, int) == [1]
    assert misc.str2iter("[1, ]", list, int) == [1]


def test_batch() -> None:
    """Test batching of iterable."""

    # batchsize of zero is not allowed
    batch = misc.batch(range(10), 0)
    with pytest.raises(ValueError):
        next(batch)

    # empty iterator returns empty batch
    batch = misc.batch(range(0), 5)
    with pytest.raises(StopIteration):
        next(batch)

    # batches are generated as expected
    iterable = range(12)
    batch = misc.batch(iterable, 5)
    assert next(batch) == range(5)
    assert next(batch) == range(5, 10)
    assert next(batch) == range(10, 12)

    # batchsize larger then length of iterator returns iterator unchanged
    iterable = range(10)
    batch = misc.batch(iterable, 12)
    assert next(batch) == range(10)


def test_wlsSamplingBound() -> None:
    """Test computation of weighted Least-Squares sampling bound."""

    # m must be positive
    with pytest.raises(AssertionError):
        _ = misc.wlsSamplingBound(m=0, c=1)
    with pytest.raises(AssertionError):
        _ = misc.wlsSamplingBound(m=-10, c=1)

    # c must be positive
    with pytest.raises(AssertionError):
        _ = misc.wlsSamplingBound(m=10, c=0)
    with pytest.raises(AssertionError):
        _ = misc.wlsSamplingBound(m=10, c=-1)

    # standard usecase
    assert misc.wlsSamplingBound(m=10, c=1) == 36
    assert misc.wlsSamplingBound(m=25, c=4) == 648


def test_paramDictToList() -> None:
    """Test conversion of pt.Parameter dictionary to index-ordered list.

    .. note::
        This function is deprecated. Tests are not needed anymore.
    """
    pass


def test_gelman_rubin_condition() -> None:
    """Test computation of Gelman-Rubin criterion."""

    # uncorrelated chains yield value close to 1
    chains = np.random.normal(0, 1, (3, 10000, 1))
    assert np.linalg.norm(misc.gelman_rubin_condition(chains)-1) < 1e-3

    # chains without variation (totally correlated) yield inf
    chains = np.ones((3, 1000, 1))
    assert np.all(misc.gelman_rubin_condition(chains) == np.inf)

    # correlated chains yield value greater than 1
    chains = np.arange(3000).reshape(3, 1000, 1)
    assert np.allclose(misc.gelman_rubin_condition(chains), 4.122986296365293)


def test_get_confidence_interval() -> None:
    """Test computation of confidence intervals."""

    smpl_uni = np.random.uniform(-1, 1, (int(1e+7), 1))

    # test symmetric mass distribution
    confidence_interval = misc.get_confidence_interval(
        smpl_uni, rate=0.5, resolution=10000)
    assert np.linalg.norm(confidence_interval-np.array([[-0.5, 0.5]])) < 1e-2

    # test rate = 0 gives median
    confidence_interval = misc.get_confidence_interval(
        smpl_uni, rate=0.0, resolution=10000)
    assert np.linalg.norm(confidence_interval-np.array([[0.0, 0.0]])) < 1e-2

    # test rate = 1 gives complete domain
    confidence_interval = misc.get_confidence_interval(
        smpl_uni, rate=1.0, resolution=10000)
    assert np.linalg.norm(confidence_interval-np.array([[-1.0, 1.0]])) < 1e-2

    # test asymmetric mass distribution
    smpl_tri = np.random.triangular(0, 1, 5, (int(1e+6), 1))

    def conf_int(r):
        """Confidence interval for triangular density with a=0, b=5 and c=1."""
        assert 0 <= r <= 1
        z2 = 5-np.sqrt(10-10*r)
        z1 = np.sqrt(5/2-5*r/2) if r > 3/5 else 5 - np.sqrt(10+10*r)
        return np.array([[z1, z2]])
    for rate in [0.1, 0.6, 0.9]:  # these are relatively nice conf intervals
        interval = misc.get_confidence_interval(
            smpl_tri, rate=rate, resolution=10_000)
        assert np.linalg.norm(conf_int(rate)-interval) < 1e-2

    # test non-scalar parameter shape
    smpl_uni_multi = smpl_uni.reshape(-1, 5)
    val = np.ones((smpl_uni_multi.shape[1], 2)) * np.array([[-0.5, 0.5]])
    interval = misc.get_confidence_interval(
        smpl_uni_multi, rate=0.5, resolution=10_000)
    assert np.linalg.norm(val-interval) < 1e-2

    # test no variation in data
    smpl_one = np.ones((1, 1))
    confidence_interval = misc.get_confidence_interval(
        smpl_one, rate=0.5, resolution=10000)
    val = np.ones((smpl_one.shape[1], 2))
    assert np.linalg.norm(confidence_interval-val) < 1e-2
    confidence_interval = misc.get_confidence_interval(
        smpl_one, rate=1.0, resolution=10000)
    assert np.linalg.norm(confidence_interval-val) < 1e-2
    confidence_interval = misc.get_confidence_interval(
        smpl_one, rate=0.0, resolution=10000)
    assert np.linalg.norm(confidence_interval-val) < 1e-2


def test_doerfler_marking() -> None:
    # empty set cannot be marked
    with pytest.raises(AssertionError):
        _ = misc.doerfler_marking([], threshold=0.9)

    # test single element set
    _, _, marker = misc.doerfler_marking([1], threshold=0.9)
    assert marker == 1

    # test truncation of zeros
    test_set = np.array([0, 0, 5, 0])
    idx, values, marker = misc.doerfler_marking(test_set, threshold=0.9)
    assert np.array_equal(test_set[idx], values)
    assert np.array_equal(values[:marker], np.array([[5]]))

    # test normal use case
    test_set = np.array([0.5, 0.3, 0.1, 0.05, 0.05])
    idx, values, marker = misc.doerfler_marking(test_set, threshold=0.85)
    assert np.array_equal(values[:marker], np.array(
        [0.5, 0.3, 0.1]).reshape(-1, 1))

    # enforce marked set is larger (strict) then threshold times test set
    test_set = np.array([0.5, 0.3, 0.1, 0.05, 0.05]).reshape(-1, 1)
    idx, values, marker = misc.doerfler_marking(test_set, threshold=0.9)
    assert np.array_equal(values[:marker], np.array(
        [0.5, 0.3, 0.1, 0.05]).reshape(-1, 1))



"""Test pythia.index module."""
# from unittest.mock import MagicMock
import numpy as np
from pythia import index


def test_IndexSet() -> None:
    """Test Multiindex class."""
    indices = np.array([[0, 1], [0, 0], [1, 1], [0, 0], [2, 0]], dtype=int)
    index_set = index.IndexSet(indices)

    # test get_sobol_tuple_list()
    expected = [(1,), (2,), (1, 2)]
    val = index_set.sobol_tuples
    assert len(expected) == len(val)
    assert np.all([np.all(left == right)
                  for left, right in zip(expected, val)])

    # test get_index_number()
    val = index_set.get_index_number(np.array([[2, 0], [0, 0]]))
    expected = np.array([3, 0], dtype=int)
    assert np.all(val == expected)
    assert np.all(index_set.indices[val] == np.array([[2, 0], [0, 0]]))

    # test get_sobol_tuple_number()
    val = index_set.get_sobol_tuple_number([(1,), (1, 2)])
    assert np.all(val == [0, 2])

    # test index_to_sobol_tuple()
    val = index_set.index_to_sobol_tuple(indices)
    expected = [(2,), (), (1, 2), (), (1,)]
    assert len(val) == len(expected)
    assert np.all([np.all(left == right)
                  for left, right in zip(expected, val)])

    # test sobol_tuple_to_indices()
    val = index_set.sobol_tuple_to_indices([(1,), (1, 2)])
    expected = [np.array([[2, 0]], dtype=int), np.array([[1, 1]], dtype=int)]
    assert np.all([np.all(v == e) for v, e in zip(val, expected)])


def test_sort_index_array() -> None:
    """Test sort_index_array."""
    indices = np.array([[0, 1], [0, 0], [1, 1], [0, 0], [2, 0]], dtype=int)
    expected = np.array([[0, 0], [0, 1], [1, 1], [2, 0]], dtype=int)
    assert np.all(index.sort_index_array(indices) == expected)


def test_add_indices() -> None:
    """Test add_indices."""
    indices_1 = np.array([[0, 0], [1, 0], [2, 0]], dtype=int)
    indices_2 = np.array([[0, 0], [0, 1], [0, 2]], dtype=int)
    expected = np.array([[0, 0], [0, 1], [1, 0], [0, 2], [2, 0]], dtype=int)
    assert np.all(index.add_indices([indices_1, indices_2]) == expected)


def test_subtract_indices() -> None:
    """Test subtract_indices."""
    indices = np.array([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=int)
    subtract = np.array([[1, 0], [2, 0]], dtype=int)
    expected = np.array([[0, 0], [0, 1], [1, 1]], dtype=int)
    val = index.subtract_indices(indices, subtract)
    assert np.all(val == expected)


def test_tensor() -> None:
    """Test tensor."""
    lower = [0, 1]
    shape = [3, 3]
    expected = np.array(
        [[0, 1], [0, 2], [1, 1], [1, 2], [2, 1], [2, 2]], dtype=int)
    assert np.all(index.tensor(shape, lower) == expected)

    shape = [1, 4, 2]
    lower = [0, 0, 1]
    expected = np.array(
        [[0, 0, 1], [0, 1, 1], [0, 2, 1], [0, 3, 1]], dtype=int)
    assert np.all(index.tensor(shape, lower) == expected)


def test_simplex() -> None:
    """Test simplex."""
    val = index.simplex(2, 2)
    expected = np.array(
        [[0, 0], [0, 1], [1, 0], [0, 2], [1, 1], [2, 0]], dtype=int)
    assert np.all(val == expected)

    val = index.simplex(3, 1)
    expected = np.array(
        [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]], dtype=int)
    assert np.all(val == expected)


if __name__ == "__main__":
    test_subtract_indices()



"""Test pythia.chaos module."""
from unittest.mock import MagicMock
import numpy as np

from pythia import chaos


def test_PolynomialChaos() -> None:
    """Test PolynomialChaos."""

    def fun_1(x): return np.array(x[:, 0]**2 + x[:, 1]**2).reshape(-1, 1)
    def fun_2(x): return np.array(2*x[:, 0]).reshape(-1, 1)

    def mock_sobol_to_indices(sdx):
        if sdx[0] == (1,):
            return [np.array([[1, 0], [2, 0]])]
        if sdx[0] == (2,):
            return [np.array([[0, 1], [0, 2]])]
        if sdx[0] == (1, 2):
            return [np.array([[1, 1]])]
        raise NotImplementedError

    def mock_index_number(mdx):
        if np.linalg.norm(mdx[0] - np.array([0, 0])) < 1e-14:
            return np.array([0], dtype=int)
        if np.linalg.norm(mdx[0] - np.array([1, 0])) < 1e-14:
            return np.array([2, 5], dtype=int)
        if np.linalg.norm(mdx[0] - np.array([0, 1])) < 1e-14:
            return np.array([1, 3], dtype=int)
        if np.linalg.norm(mdx[0] - np.array([1, 1])) < 1e-14:
            return np.array([4], dtype=int)
        raise NotImplementedError

    param = MagicMock()
    param.domain = [-1, 1]
    param.distribution = "uniform"

    index_set = MagicMock()
    index_set.indices = np.array(
        [[0, 0], [0, 1], [1, 0], [0, 2], [1, 1], [2, 0]])
    index_set.sobol_tuples = [(1,), (2,), (1, 2)]
    index_set.max = np.array([2, 2], dtype=int)
    index_set.sobol_tuple_to_indices = MagicMock(
        side_effect=mock_sobol_to_indices)
    index_set.get_index_number = MagicMock(side_effect=mock_index_number)

    x_train = np.random.uniform(-1, 1, (100_000, 2))
    w_train = 1 / x_train.shape[0] * np.ones(x_train.shape[0])
    y_train = fun_1(x_train)

    pc = chaos.PolynomialChaos(
        [param, param], index_set, x_train, w_train, y_train)

    # test basic properties
    assert isinstance(pc, chaos.PolynomialChaos)
    assert np.linalg.cond(pc.gramian) < 1.1
    eye = np.eye(pc.gramian.shape[0])
    assert np.linalg.norm(pc.gramian-eye) < 1e-1

    # test approximation
    x_test = np.random.uniform(-1, 1, (1000, 2))
    assert np.linalg.norm(
        fun_1(x_test)-pc.eval(x_test), axis=0)[0] < 1e-10
    assert np.linalg.norm(
        fun_2(x_test)-pc.eval(x_test, partial=[1, 0]), axis=0)[0] < 1e-10

    # test mean, var and std
    assert np.linalg.norm(pc.mean-2/3) < 1e-10
    assert np.linalg.norm(pc.var-(8/45)) < 1e-10
    assert np.linalg.norm(pc.std-np.sqrt(8/45)) < 1e-10

    # test sobol indices
    assert np.linalg.norm(pc.sobol - np.array([[0.5], [0.5], [0]])) < 1e-10

    # test multi dimensional function output shape
    y_train = np.concatenate([fun_1(x_train)]*3, axis=1)
    pc = chaos.PolynomialChaos(
        [param, param], index_set, x_train, w_train, y_train)
    assert pc.coefficients.ndim == 2 and pc.coefficients.shape[1] == 3
    assert np.linalg.norm(pc.coefficients-pc.coefficients[:, :1]) < 1e-14
    assert pc.eval(x_train).shape == y_train.shape
    assert pc.sobol.shape == (3, 3)


def test_find_optimal_indices() -> None:
    """Test find_optimal_indices."""
    def fun_1(x):
        return np.array(x[:, 0]**2 + 1e-02*x[:, 1]**2).reshape(-1, 1)

    param = MagicMock()
    param.domain = [-1, 1]
    param.distribution = "uniform"

    x_train = np.random.uniform(-1, 1, (1000, 2))
    w_train = np.ones(x_train.shape[0]) / x_train.shape[0]
    y_train = fun_1(x_train)
    max_terms = 25

    indices, sobol = chaos.find_optimal_indices(
        [param, param], x_train, w_train, y_train, max_terms, threshold=1e-03)
    expected = np.array([np.arange(25), np.zeros(25)], dtype=int).T
    assert np.all(indices == expected)
    assert sobol[0, 0] > 1e-03
    assert sobol[1, 0] < 1e-03
    assert sobol[2, 0] < 1e-10

    indices, sobol = chaos.find_optimal_indices(
        [param, param], x_train, w_train, y_train,
        max_terms=10, threshold=1e-06)
    expected = np.array([[0, 0], [0, 1], [1, 0], [0, 2], [2, 0],
                         [0, 3], [3, 0], [0, 4], [4, 0]])
    assert np.all(indices == expected)
    assert sobol[0, 0] > 1e-06
    assert sobol[1, 0] > 1e-06
    assert sobol[2, 0] < 1e-10


def test_assemble_indices() -> None:
    """Test assemble_indices."""
    sobol_tuples = [(1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)]
    enum_idx = [0, 3, 2, 6]
    max_terms = 200
    val = chaos.assemble_indices(enum_idx, sobol_tuples, max_terms)
    assert val.shape[0] < max_terms
    assert val.shape[1] == 3
    assert [0, 0, 1] in val.tolist()
    assert [1, 1, 0] in val.tolist()
    assert [0, 1, 0] not in val.tolist()
    assert [1, 0, 1] not in val.tolist()


def test_get_gram_batchsize() -> None:
    """Test infering batch size for Gram matrix."""
    # Don't know how to test this.
    pass


if __name__ == "__main__":
    test_PolynomialChaos()
    test_find_optimal_indices()



""" Test pythia.basis module. """
from unittest.mock import MagicMock
import pytest
import numpy as np
import scipy.stats
from scipy.integrate import quad
from scipy.special import gamma, factorial, legendre, jacobi, laguerre

import pythia.basis as basis


def test_set_legendre_basis() -> None:
    """Test assembling of shifted Legendre basis functions."""
    # test if Hermite basis is orthogonal with correct scaling
    deg = 5
    param = MagicMock()
    param.domain = [-1, 2]
    base = basis.set_legendre_basis(param, deg)
    assert len(base) == deg+1
    expected = np.diag([(param.domain[1]-param.domain[0])/(2*j+1)
                       for j in range(deg+1)])
    gramian = np.zeros((deg+1, deg+1))
    for row, p in enumerate(base):
        for col, q in enumerate(base):
            def integrand(y):
                return p(y) * q(y)
            gramian[row, col] = quad(integrand, *param.domain)[0]
    assert np.linalg.norm(expected-gramian) < 1e-10


def test_set_hermite_basis() -> None:
    """Test assembling of scaled Hermite basis functions."""
    # test if missing (optional) parameter attributes are detected
    param = MagicMock()
    param.domain = [-np.inf, np.inf]
    param.mean = 1
    with pytest.raises(AssertionError):
        _ = basis.set_hermite_basis(param, 5)
    param.var = "2"
    with pytest.raises(AssertionError):
        _ = basis.set_jacobi_basis(param, 5)

    # test if Hermite basis is orthogonal with correct scaling
    deg = 5
    param = MagicMock()
    param.domain = [-np.inf, np.inf]
    param.mean = 0.5
    param.var = 1.75
    base = basis.set_hermite_basis(param, deg)
    assert len(base) == deg+1
    expected = np.diag([np.sqrt(2*np.pi) * factorial(j)
                       for j in range(deg+1)])
    gramian = np.zeros((deg+1, deg+1))
    for row, p in enumerate(base):
        for col, q in enumerate(base):
            def integrand(y):
                var = param.var
                pdf = 1 / np.sqrt(var) * np.exp(-(y-param.mean)**2 / (2*var))
                return p(y) * q(y) * pdf
            gramian[row, col] = quad(integrand, *param.domain)[0]
    assert np.linalg.norm(expected-gramian) < 1e-10


def test_set_jacobi_basis() -> None:
    """Test assembling of shifted Jacobi basis functions.

    The standard Jacobi polynomials shifted to the interval :math:`[a,b]`
    satisfy the orthogonality relation

    .. math::
        \\int_a^b P_m^{(\\beta-1,\\alpha-1)}(y) P_n^{(\\beta-1,\\alpha-1)}(y) (y-a)^{\\alpha-1} (b-y)^{\\beta-1}\\,\\mathrm{d}y
        =  \\frac{(b-a)^{\\alpha+\\beta-1}}{2m+\\alpha+\\beta-1} \\frac{1}{m!} \\frac{\\Gamma(m+\\alpha)\\Gamma(m+\\beta)}{\\Gamma(m+\\alpha+\\beta-1)} \\delta_{mn}.
    """
    # test if missing (optional) parameter attributes are detected
    param = MagicMock()
    param.domain = [-2, 7]
    param.alpha = 1
    with pytest.raises(AssertionError):
        _ = basis.set_jacobi_basis(param, 5)
    param.beta = "2"
    with pytest.raises(AssertionError):
        _ = basis.set_jacobi_basis(param, 5)

    # test if Jacobi basis is orthogonal with correct scaling
    def factor(m, alpha, beta):
        """Compute the scaling of the inner product of Jacobi polynomials."""
        width = param.domain[1] - param.domain[0]
        numerator = width**(alpha+beta-1) * gamma(m+alpha) * gamma(m+beta)
        denominator = ((2*m+alpha+beta-1) * factorial(m)
                       * gamma(m+alpha+beta-1))
        # Note: The Jacobi polynomials created by PyThia have a normalized
        #       leading coefficient. This is why the scalar product needs to be
        #       scaled accordingly by the leading coefficient of the standard
        #       Jacobi polynomials and the m-th power of the leading
        #       coefficient of the shift :math:` y \\mapsto 2/(b-a) (y-a) - 1`.
        lead_coeff = jacobi(m, beta-1, alpha-1).coeffs[0]
        normalization_scale = ((param.domain[1]-param.domain[0])**(2*m)
                               / (2**(2*m) * lead_coeff**2))
        return normalization_scale * numerator / denominator
    deg = 5
    param = MagicMock()
    param.domain = [1, 4]
    param.alpha = 1
    param.beta = 2
    base = basis.set_jacobi_basis(param, deg)
    assert len(base) == deg+1
    expected = np.diag([factor(j, param.alpha, param.beta)
                       for j in range(deg+1)])
    gramian = np.zeros((deg+1, deg+1))
    for row, p in enumerate(base):
        for col, q in enumerate(base):
            def integrand(y):
                a = param.domain[0]
                b = param.domain[1]
                pdf = (y-a)**(param.alpha-1) * (b-y)**(param.beta-1)
                return p(y) * q(y) * pdf
            gramian[row, col] = quad(integrand, *param.domain)[0]
    assert np.linalg.norm(expected-gramian) < 1e-10


def test_set_laguerre_basis() -> None:
    """Test assembling of shifted and scaled Laguerre basis functions.

    The standard Laguerre polynomials scaled with :math:`\\beta` and shifted to
    the interval :math:`(a,\\infty)` satisfy the orthogonality relation

    .. math::
        \\int_a^\\infty P_m^{\\alpha-1}(y) P_n^{\\alpha-1}(y) (y-a)^{\\alpha-1} \\exp(-\\beta(y-a))\\,\\mathrm{d}y
        = \\frac{\\Gamma(m+\\alpha)}{\\beta^{\\alpha} m!} \\delta_{mn}.
    """

    # test if missing (optional) parameter attributes are detected
    param = MagicMock()
    param.domain = [1, np.inf]
    param.alpha = 1
    with pytest.raises(AssertionError):
        _ = basis.set_laguerre_basis(param, 5)
    param.beta = "2"
    with pytest.raises(AssertionError):
        _ = basis.set_laguerre_basis(param, 5)

    # test if Laguerre basis is orthogonal with correct scaling
    def factor(m, alpha, beta):
        """Compute the scaling of the inner product of Laguerre polynomials."""
        numerator = gamma(m+alpha)
        denominator = beta**alpha * factorial(m)
        # Note: The Laguerre polynomials created by PyThia have a normalized
        #       leading coefficient. This is why the scalar product needs to be
        #       scaled accordingly by the leading coefficient of the standard
        #       Laguerre polynomials and the m-th power of the leading
        #       coefficient of the shift :math:`y \\mapsto \\beta (y-a)`.
        lead_coeff = laguerre(m, alpha-1).coeffs[0]
        normalization_scale = 1 / (beta**(2*m) * lead_coeff**2)
        return normalization_scale * numerator / denominator
    deg = 5
    param = MagicMock()
    param.domain = [1, np.inf]
    param.alpha = 1.0
    param.beta = 2.0
    base = basis.set_laguerre_basis(param, deg)
    assert len(base) == deg+1
    expected = np.diag([factor(j, param.alpha, param.beta)
                       for j in range(deg+1)])
    gramian = np.zeros((deg+1, deg+1))
    for row, p in enumerate(base):
        for col, q in enumerate(base):
            def integrand(y):
                a = param.domain[0]
                w = (y-a)**(param.alpha-1) * np.exp(-param.beta*(y-a))
                return p(y) * q(y) * w
            gramian[row, col] = quad(integrand, *param.domain)[0]
    assert np.linalg.norm(expected-gramian) < 1e-10


def test_normalize_polynomial() -> None:
    """Test normalization of polynomial bases."""
    deg = 5

    # normalization of Legendre basis
    def w_uni(y):
        return scipy.stats.uniform.pdf(
            y, loc=param.domain[0], scale=param.domain[1]-param.domain[0])
    param = MagicMock()
    param.domain = [-1, 2]
    base = basis.set_legendre_basis(param, deg)
    base = basis.normalize_polynomial(w_uni, base, param)
    gramian = np.zeros(deg+1)
    for j, poly in enumerate(base):
        def int_uni(y):
            return poly(y)**2 * w_uni(y)
        gramian[j] = quad(int_uni, *param.domain)[0]
    assert np.linalg.norm(gramian - np.ones(deg+1)) < 1e-10

    # normalization of Hermite basis
    def w_norm(y):
        return scipy.stats.norm.pdf(
            y, loc=param.mean, scale=np.sqrt(param.var))
    param = MagicMock()
    param.domain = [-np.inf, np.inf]
    param.mean = 1
    param.var = 2
    base = basis.set_hermite_basis(param, deg)
    base = basis.normalize_polynomial(w_norm, base, param)
    gramian = np.zeros(deg+1)
    for j, poly in enumerate(base):
        def int_norm(y):
            return poly(y)**2 * w_norm(y)
        gramian[j] = quad(int_norm, *param.domain)[0]
    assert np.linalg.norm(gramian - np.ones(deg+1)) < 1e-10

    # normalization of Laguerre basis
    def w_gamma(y):
        val = y - param.domain[0]
        return scipy.stats.gamma.pdf(val, a=param.alpha, scale=1/param.beta)
    param = MagicMock()
    param.domain = [1, np.inf]
    param.alpha = 1
    param.beta = 2
    base = basis.set_laguerre_basis(param, deg)
    base = basis.normalize_polynomial(w_gamma, base, param)
    gramian = np.zeros(deg+1)
    for j, poly in enumerate(base):
        def int_gamma(y):
            return poly(y)**2 * w_gamma(y)
        gramian[j] = quad(int_gamma, *param.domain)[0]
    assert np.linalg.norm(gramian - np.ones(deg+1)) < 1e-10

    # normalization of Jacobi basis
    def w_beta(y):
        a, b = param.domain
        val = (y-a) / (b-a)
        return scipy.stats.beta.pdf(val, a=param.alpha, b=param.beta)/(b-a)
    param = MagicMock()
    param.domain = [1, 5]
    param.alpha = 2
    param.beta = 0.5
    base = basis.set_jacobi_basis(param, deg)
    base = basis.normalize_polynomial(w_beta, base, param)
    gramian = np.zeros(deg+1)
    for j, poly in enumerate(base):
        def int_beta(y):
            return poly(y)**2 * w_beta(y)
        gramian[j] = quad(int_beta, *param.domain)[0]
    assert np.linalg.norm(gramian - np.ones(deg+1)) < 1e-10


def test_univariate_basis() -> None:
    """Test creation of univariate normalized polynomial basis."""

    # test correct assign of basis polynomials
    def weight(y, param):
        if param.distribution == "uniform":
            return scipy.stats.uniform.pdf(
                y, loc=param.domain[0], scale=param.domain[1]-param.domain[0])
        # return normal pdf
        return scipy.stats.norm.pdf(
            y, loc=param.mean, scale=np.sqrt(param.var))
    deg = 5
    uni_param = MagicMock()
    uni_param.domain = [0, 1]
    uni_param.distribution = "uniform"
    norm_param = MagicMock()
    norm_param.domain = [-np.inf, np.inf]
    norm_param.mean = 0
    norm_param.var = 1
    norm_param.distribution = "normal"
    params = [uni_param, norm_param]
    base = basis.univariate_basis(params, [deg]*len(params))
    for param, polys in zip(params, base):
        gramian = np.zeros(deg+1)
        for j, poly in enumerate(polys):
            def integrand(y):
                return poly(y)**2 * weight(y, param)
            gramian[j] = quad(integrand, *param.domain)[0]
        assert np.linalg.norm(gramian - np.ones(deg+1)) < 1e-10


def test_multivariate_basis() -> None:
    """Test creation of multivariate normalized polynomial basis."""

    param_basis = [[legendre(j) for j in range(3)]]*2
    mdx = np.array([[0, 0], [1, 0], [0, 1], [0, 2]])

    # test handling of shape mismatch
    with pytest.raises(AssertionError):
        _ = basis.multivariate_basis(param_basis, mdx[:, :1])
    with pytest.raises(AssertionError):
        _ = basis.multivariate_basis(param_basis, mdx, partial=[1, 0, 0])

    # test computation of basis
    base = basis.multivariate_basis(param_basis, mdx)
    for j, idx in enumerate(mdx):
        y = np.random.uniform(-1, 1, (1, 2))
        expected = (param_basis[0][idx[0]](y[0, 0])
                    * param_basis[1][idx[1]](y[0, 1]))
        assert np.linalg.norm(base[j](y) - expected) <= 1e-14

    # test computation of partial derivatives
    partial = [0, 1]
    base2 = basis.multivariate_basis(param_basis, mdx, partial)
    assert np.all(np.abs(base2[0](np.random.uniform(-1, 1, 100)) < 1e-14))
    ys = np.random.uniform(-1, 1, 200).reshape(100, 2)
    assert np.all(np.abs(base2[-1](ys) - 3*base[-2](ys)) < 1e-14)
